#!/usr/bin/ruby -w
# -*- ruby -*-

#-------------------------------------------------------
# ANSIColor module
#-------------------------------------------------------

module ANSIColor

  # Adds the escape sequence for value retrieval.
  class ANSIEscapedColors < Hash
    def [](str)
      val = super
      if val
        val = "\e[" + val + "m"
      end
      val
    end
  end

  @@ATTRIBUTE_COLORS = ANSIEscapedColors[
    'none'       => '00', 
    'bold'       => '01',
    'underscore' => '04',
    'underline'  => '04',
    'blink'      => '05',
    'reverse'    => '07',
    'concealed'  => '08'
  ]

  @@TEXT_COLORS = ANSIEscapedColors[
    'black'   => '30',
    'red'     => '31',
    'green'   => '32',
    'yellow'  => '33',
    'blue'    => '34',
    'magenta' => '35',
    'cyan'    => '36',
    'white'   => '37'
  ]
  
  @@BACKGROUND_COLORS = ANSIEscapedColors[
    'black'   => '40',
    'red'     => '41',
    'green'   => '42',
    'yellow'  => '43',
    'blue'    => '44',
    'magenta' => '45',
    'cyan'    => '46',
    'white'   => '47'
  ]

  @@RESET = "\e[0m"

  # returns the code for the given color string, which is in the format:
  # [foreground] on [background]. Note that the foreground and background sections
  # can have modifiers (attributes). Examples:
  #     black
  #     blue on white
  #     bold green on yellow
  #     underscore bold magenta on cyan
  #     underscore red on bold cyan

  def ANSIColor.code(str)
    fg, bg = str.split(/\s*\bon_?\s*/)
    foreground(fg) + background(bg)
  end

  # returns the code for the given background color(s)
  def ANSIColor.background(bgcolor)
    make_code(bgcolor, [ @@BACKGROUND_COLORS ])
  end

  # returns the code for the given foreground color(s)
  def ANSIColor.foreground(fgcolor)
    make_code(fgcolor, [ @@ATTRIBUTE_COLORS, @@TEXT_COLORS ])
  end

  def ANSIColor.reset
    @@RESET
  end

  protected

  def ANSIColor.make_code(str, colors)
    code = ""
    return code unless str
    str.split.each do |s|
      found = false
      colors.each do |ca|
        if c = ca[s]
          code += c 
          found = true
          break
        end
      end
      unless found
        $stderr.puts "WARNING: ANSIColor::make_code(" + str + "): unknown color: " + s
        break
      end
    end
    code
  end

end
