#!/usr/bin/ruby -w
# -*- ruby -*-

require './log'
require './glarkfile'


# A function object, which can be applied (processed) against a InputFile.
class FuncObj
  
  attr_accessor :match_line_number, :file, :matches

  def initialize
    @match_line_number = nil
    @matches = Array.new
  end

  def add_match(lnum)
    @matches.push(lnum)
  end

  def start_position
    match_line_number
  end

  def end_position
    start_position
  end

  def reset_file(file)
    @match_line_number = nil
    @file              = file
    @matches           = Array.new
  end

  def range(var, count)
    if var
      if var.index(/([\.\d]+)%/)
        count * $1.to_f / 100
      else
        var.to_f
      end
    else
      nil
    end
  end

  def process(infile)
    got_match = false
    reset_file(infile.fname)

    rgstart = range($options.range_start, infile.lines.size)
    rgend = range($options.range_end, infile.lines.size)

    # puts "rgstart = #{rgstart}"
    # puts "rgend = #{rgend}"

    nmatches = 0
    (0 ... infile.lines.size).each do |lnum|
      #puts "start #{lnum} >= #{rgstart}: #{lnum >= rgstart}"
      #puts "end #{lnum} <= #{rgend}: #{lnum <= rgend}"

      # puts "running at lnum #{lnum}"

      if (!rgstart || lnum >= rgstart) && 
          (!rgend || lnum <= rgend) &&
          evaluate(infile.lines[lnum], lnum, infile.fname)
        mark_as_match(infile)
        got_match = true
        nmatches += 1
        break if $options.num_matches && nmatches >= $options.num_matches
      end
    end
    
    if $options.file_names_only
      if got_match != $options.invert_match
        print infile.fname
        if $options.write_null
          print "\0"
        else
          print "\n"
        end
      end
    elsif $options.invert_match
      infile.write_non_matches
    elsif got_match
      infile.write_matches
    end
  end

  def mark_as_match(infile)
    infile.mark_as_match(start_position)
  end
  
  def to_s
    return inspect
  end
  
end


# Applies a regular expression against a InputFile.
class RegexpFuncObj < FuncObj

  attr_reader :re

  def initialize(re)
    @re = re
    @file = nil
    super()
  end

  def inspect
    @re.inspect
  end

  def evaluate(line, lnum, file)
    # spew self, "evaluating <<<#{line[0 .. -2]}>>>"
    if md = @re.match(line)
      # spew self, "matched regular expression #{@re}"

      if $options.extract_matches
        Log.log "replacing line"
        line.replace(md[-1] + "\n")
        # line.gsub!(@re) { |m| Log.log "replacing with #{m}"; m }
      else
        Log.log "NOT replacing line"
      end
      
      @match_line_number = lnum
      # highlight what the regular expression matched
      if $options.highlight
        # must use the block form
        line.gsub!(@re) { |m| $options.text_highlight + m + ANSIColor.reset }
      end
      add_match(lnum)
      return true
    else
      return false
    end
  end
  
end


# Associates a pair of expressions.
class CompoundExpression < FuncObj

  def initialize(op1, op2)
    @op1, @op2 = op1, op2
    @file = nil
    super()
  end

  def reset_file(file)
    @op1.reset_file(file)
    @op2.reset_file(file)
    super
  end

  def start_position
    return @last_start
  end
  
end


# Evaluates both expressions, and is satisfied when either return true.
class OrExpression < CompoundExpression

  def evaluate(line, lnum, file)
    # spew self, "evaluating <<<#{line[0 .. -2]}>>>"

    m1 = @op1.evaluate(line, lnum, file)
    m2 = @op2.evaluate(line, lnum, file)

    if m1 or m2
      if m1
        @last_start = @op1.start_position
        @last_end   = @op1.end_position
      end
      if m2
        @last_start = @op2.start_position
        @last_end   = @op2.end_position
      end
      
      @match_line_number = lnum
      add_match(lnum)
      return true
    else
      return false
    end
  end
  
  def inspect
    "(" + @op1.to_s + " or " + @op2.to_s + ")"
  end

  def end_position
    return @last_end
  end

end


# Evaluates both expressions, and is satisfied when both return true.
class AndExpression < CompoundExpression
  
  def initialize(dist, op1, op2)
    @dist = dist
    super(op1, op2)
  end

  def mark_as_match(infile)
    infile.mark_as_match(start_position, end_position)
  end

  def match_within_distance(op, lnum)
    op.matches.size > 0 and (op.matches[-1] - lnum <= @dist)
  end

  def inspect
    str = "("+ @op1.to_s
    if @dist == 0
      str += " same line as "
    elsif @dist == $options.infinite_distance
      str += " same file as "
    else 
      str += " within " + @dist.to_s + " lines of "
    end
    str += @op2.to_s + ")"
    str
  end

  def reset_match(op, lnum)
    op.matches.reverse.each do |m|
      if lnum - m <= @dist
        @last_start = m
        return true
      else
        return false
      end
    end
    return true
  end

  def match?(line, lnum, file)
    m1 = @op1.evaluate(line, lnum, file)
    m2 = @op2.evaluate(line, lnum, file)

    if m1 and match_within_distance(@op2, lnum)
      return reset_match(@op2, lnum)
    elsif m2 and match_within_distance(@op1, lnum)
      return reset_match(@op1, lnum)
    else
      return false
    end
  end

  def end_position
    [@op1.end_position, @op2.end_position].max
  end

  def evaluate(line, lnum, file)
    # spew self, "evaluating line #{lnum}: #{line[0 .. -2]}"

    if match?(line, lnum, file)
      @match_line_number = lnum
      return true
    else
      return false
    end
  end
end


class ExpressionCreator

  attr_reader :expr

  def initialize(arg, args)
    @current = arg
    @args    = args
    @expr    = create_expression
  end

  def create_regular_expression
    pat = @current

    # this check is because they may have omitted the pattern, e.g.:
    #   % glark *.cpp
    if File.exists?(pat) and !$options.quiet
      $stderr.print "WARNING: pattern '#{pat}' exists as a file.\n"
      $stderr.print "    Pattern may have been omitted.\n"
    end

    ignorecase = $options.nocase
    
    # we handle a ridiculous number of possibilities here:
    #     /foobar/     -- "foobar"
    #     /foo/bar/    -- "foo", then slash, then "bar"
    #     /foo\/bar/   -- same as above
    #     /foo/bar/i   -- same as above, case insensitive
    #     /foo/bari    -- "/foo/bari" exactly
    #     /foo/bar\/i  -- "/foo/bar/i" exactly
    #     foo/bar/     -- "foo/bar/" exactly
    #     foo/bar/     -- "foo/bar/" exactly

    if pat.index(/^\/(.*[^\\])\/i$/)
      pattern    = $1
      ignorecase = true
    elsif pat.index(/^\/(.*[^\\])\/$/)
      pattern    = $1
    elsif pat.index(/^(\/.*)$/)
      pattern    = $1
    elsif pat.index(/^(.*\/)$/)
      pattern    = $1
    else
      pattern    = pat
    end
    
    if $options.whole_words
      # match "\w", A-Za-z0-9_, 
      stword = pattern.index(/^[\[\(]*(?:\\w|\w)/)

      endword = pattern[-1, 1] =~ /\w/ || pattern[-2, 2] =~ /\w\?\*/

      re = Regexp.new('(?:                 # one of the following:
                           \\w             #   - \w for regexp
                         |                 # 
                           \w              #   - a literal A-Z, a-z, 0-9, or _
                         |                 # 
                           (?:             #   - one of the following:
                               \[[^\]]*    #         LB, with no RB until:
                               (?:         #      - either of:
                                   \\w     #         - "\w"
                                 |         # 
                                   \w      #         - a literal A-Z, a-z, 0-9, or _
                               )           #      
                               [^\]]*\]    #      - anything (except RB) to the next RB
                           )               #
                       )                   #
                       (?:                 # optionally, one of the following:
                           \*              #   - "*"
                         |                 # 
                           \+              #   - "+"
                         |                 #
                           \?              #   - "?"
                         |                 #
                           \{\d*,\d*\}     #   - "{3,4}", "{,4}, "{,123}" (also matches the invalid {,})
                       )?                  #
                       $                   # the end.
                      ', 
                      Regexp::EXTENDED)
      endword = pattern.index(re)

      if stword && endword
        # good
      else
        msg = "WARNING: pattern '#{pattern}' does not "
        if stword
          msg += "end"
        elsif endword
          msg += "begin"
        else
          msg += "begin and end"
        end
        msg += " on a word boundary."
        $stderr.puts msg
      end
      pattern = '\b' + pattern + '\b'
    elsif $options.whole_lines
      pattern = '^'  + pattern + '$'
    end
    
    # spew "pattern", pattern
    # spew "ignorecase", ignorecase
    
    if ignorecase
      regex = Regexp.new(pattern, Regexp::IGNORECASE)
    else
      regex = Regexp.new(pattern)
    end
    
    return RegexpFuncObj.new(regex)
  end 

  # creates two expressions and returns them.
  def create_expressions
    @current = @args.shift
    a1 = create_expression

    @current = @args.shift
    a2 = create_expression
    
    [ a1, a2 ]
  end

  def consume_or_expression
    a1, a2 = create_expressions
    unless a1 && a2
      $stderr.print "ERROR: 'or' expression takes two arguments\n"
      exit 2
    end

    # explicit end tag is optional:
    @args.shift if @args[0] == "--end-of-or"
    OrExpression.new(a1, a2)
  end

  def consume_and_expression
    dist = @args.shift
    # check to ensure that this is numeric
    if !dist || (dist.to_i != $options.infinite_distance && !dist.index(/^\d+$/))
      $stderr.print "ERROR: invalid distance for 'and' expression: '#{dist}'\n" 
      $stderr.print "    expecting an integer, or #{$options.infinite_distance} for 'infinite'\n" 
      exit 2
    end

    if dist.to_i == $options.infinite_distance
      dist = 1.0 / 0.0            # infinity
    else
      dist = dist.to_i
    end

    a1, a2 = create_expressions
    unless a1 && a2
      $stderr.print "ERROR: 'and' expression takes two arguments\n"
      exit 2
    end
    # explicit end tag is optional:
    @args.shift if @args[0] == "--end-of-and"
    AndExpression.new(dist, a1, a2)
  end

  def create_expression
    if @current
      Log.log "processing arg #{@current}"
      case @current
      when "--or", "-o"
        return consume_or_expression
      when "--and", "-a"
        return consume_and_expression
      when /^--/
        $stderr.print "option not understood: #{@current}"
        exit 2
      else
        $stderr.print "assuming the last argument #{@current} is a pattern\n" if $options.verbose
        return create_regular_expression
      end
    else
      return nil
    end
  end

end


if __FILE__ == $0
  Log.verbose = true
  require './glarkopt'
  GlarkOptions.new
  ec = ExpressionCreator.new("--and", [ "4", "foo", "/bar" ])
  puts ec
  puts ec.expr
end
