#!/usr/bin/ruby -w
# -*- ruby -*-

require './log'

# A thing that can be grepped.
class InputFile

  attr_reader :lines, :fname, :stati, :count
  
  def initialize(fname, lines)
    @fname = fname
    @lines = lines
    @stati = Array.new          # index = line number, value = context character
    if $options.count
      @count = 0
    end
    @output = $options.grep_output ? GrepOutputFormat.new(self) : GlarkOutputFormat.new(self)
  end

  # prints the line, and adjusts for the fact that in our world, lines are
  # 0-indexed, whereas they are displayed as if 1-indexed.
  def print_line(lnum, ch = nil)
    if lnum < lines.size
      if $options.show_line_numbers
        printf "%5d ", lnum + 1
      end
      if ch && has_context
        printf "%s ", ch
      end
      print lines[lnum]
    else
      Log.log "line number #{lnum} exceeds line count #{lines.size}"
    end
  end

  def has_context
    $options.after != 0 || $options.before != 0
  end

  def set_status(from, to, ch, force = false)
    from.upto(to) do |ln|
      if force or not @stati[ln]
        @stati[ln] = ch
      end
    end
  end

  def mark_as_match(start_line, end_line = start_line)
    if $options.grep_output
      end_line = start_line
    end

    $options.exit_status = $options.invert_match ? 1 : 0

    if $options.count
      @count += 1
    else
      st = [0, start_line - $options.before].max
      set_status(st,           start_line - 1,    "-")
      set_status(start_line,   end_line,          ":",  true)
      set_status(end_line + 1, end_line + $options.after, "+")
    end
  end

  def write_matches
    @output.write_matches
  end

  def write_non_matches
    @output.write_non_matches
  end

end


class OutputFormat

  def initialize(infile)
    @infile = infile
  end

  # prints the line, and adjusts for the fact that in our world, lines are
  # 0-indexed, whereas they are displayed as if 1-indexed.
  def print_line(lnum, ch = nil)
    if lnum < @infile.lines.size
      if $options.show_line_numbers
        printf "%5d ", lnum + 1
      end
      if ch && has_context
        printf "%s ", ch
      end
      print @infile.lines[lnum]
    end
  end

  def has_context
    $options.after != 0 || $options.before != 0
  end

  def write_matches
    if $options.count
      write_count
    else
      lastln = nil
      0.upto(@infile.stati.size) do |ln|
        if @infile.stati[ln]
          print_line(ln, @infile.stati[ln]) 
          lastln = ln
        end
      end
    end
  end

  def write_non_matches
    if $options.count
      write_non_count
    else
      (0 ... @infile.lines.length).each do |ln|
        unless @infile.stati[ln] && @infile.stati[ln] == ":"
          print_line(ln) 
        end
      end
    end
  end

end


class GlarkOutputFormat < OutputFormat

  def show_file_header
    if $options.show_file_names
      print $options.file_highlight if $options.highlight
      print @infile.fname, ":"
      print ANSIColor.reset if $options.highlight
      print "\n"
    end
  end

  def write_count
    puts "    " + @infile.count.to_s
  end

  def write_non_count
    puts "    " + (@infile.lines.size - @infile.count).to_s
  end

  def write_matches
    show_file_header if ARGV.size > 1
    super
  end

  def write_non_matches
    show_file_header if ARGV.size > 1
    super
  end

end


# This matches grep, mostly. It is for running within emacs, thus,
# it does not support context or highlighting.
class GrepOutputFormat < OutputFormat

  def initialize(infile)
    @show_file_name = ARGV.size > 1 && $options.show_file_names
    super
  end

  def write_count
    print @infile.fname, ":" if @show_file_name
    puts @infile.count
  end

  def write_non_count
    print @infile.fname, ":" if @show_file_name
    puts @infile.lines.length - @infile.count
  end

  # prints the line, and adjusts for the fact that in our world, lines are
  # 0-indexed, whereas they are displayed as if 1-indexed.
  def print_line(lnum, ch = nil)
    print @infile.fname, ":" if @show_file_name
    if $options.show_line_numbers
      printf "%d: ", lnum + 1
    end
    print @infile.lines[lnum]
  end

end


class BinaryFile < InputFile

  def write_matches
    if $options.count
      write_count
    else
      puts "Binary file " + @fname + " matches"
    end
  end

  def write_non_matches
    if $options.count
      write_non_count
    else
      puts "Binary file " + @fname + " matches"
    end
  end

end


if __FILE__ == $0
  Log.verbose = true
  require './glarkfile'
  require './glarkopt'
  go = GlarkOptions.new
  go.count = true

  f = InputFile.new("glarkformat", [ "foo\n", "bar\n" ])
  f.mark_as_match(0, 14)
  puts "matches:"
  f.write_matches
  puts "end of matches."

  go.grep_output = true

  f = InputFile.new("grepformat", %w{ foo bar baz })
  f.mark_as_match(0, 14)
  puts "matches:"
  f.write_matches
  puts "end of matches."

end
